@@ -237,14 +237,15 @@ public:
   nameExp(position pos, string s)
     : exp(pos), value(new simpleName(pos, symbol::trans(s))) {}
 
-  void prettyprint(ostream &out, Int indent);
+  void prettyprint(ostream &out, Int indent) override;
+  void createSymMap(AsymptoteLsp::SymbolContext* symContext) override;
 
-  symbol getName()
+  symbol getName() override
   {
     return value->getName();
   }
 
-  void transAsType(coenv &e, types::ty *target) {
+  void transAsType(coenv &e, types::ty *target) override {
     value->varTrans(trans::READ, e, target);
 
     // After translation, the cached type is no longer needed and should be
@@ -254,7 +255,7 @@ public:
     ct=0;
   }
 
-  types::ty *trans(coenv &e) {
+  types::ty *trans(coenv &e) override {
     types::ty *t=cgetType(e);
     if (t->kind == types::ty_error) {
       em.error(getPos());
@@ -272,32 +273,32 @@ public:
     }
   }
 
-  types::ty *getType(coenv &e) {
+  types::ty *getType(coenv &e) override {
     types::ty *t=value->varGetType(e);
     return t ? t : types::primError();
   }
 
-  trans::varEntry *getCallee(coenv &e, types::signature *sig) {
+  trans::varEntry *getCallee(coenv &e, types::signature *sig) override {
 #ifdef DEBUG_GETAPP
     cout << "nameExp" << endl;
 #endif
     return value->getCallee(e, sig);
   }
 
-  void transWrite(coenv &e, types::ty *target, exp *newValue) {
+  void transWrite(coenv &e, types::ty *target, exp *newValue) override {
     newValue->transToType(e, target);
     this->value->varTrans(trans::WRITE, e, target);
 
     ct=0;  // See note in transAsType.
   }
 
-  void transCall(coenv &e, types::ty *target) {
+  void transCall(coenv &e, types::ty *target) override {
     value->varTrans(trans::CALL, e, target);
 
     ct=0;  // See note in transAsType.
   }
 
-  exp *evaluate(coenv &, types::ty *) {
+  exp *evaluate(coenv &, types::ty *) override {
     // Names have no side-effects.
     return this;
   }
@@ -363,9 +364,15 @@ class fieldExp : public nameExp {
     void print(ostream& out) const {
       out << "<exp>";
     }
-    symbol getName() {
+
+    symbol getName() const {
       return object->getName();
     }
+
+    AsymptoteLsp::SymbolLit getLit() const
+    {
+      return AsymptoteLsp::SymbolLit(static_cast<std::string>(object->getName()));
+    }
   };
 
   // Try to get this into qualifiedName somehow.
@@ -513,6 +520,13 @@ public:
 
   types::ty *trans(coenv &e);
   types::ty *getType(coenv &) { return types::primInt(); }
+
+  template<typename T>
+  [[nodiscard]]
+  T getValue() const
+  {
+    return static_cast<T>(value);
+  }
 };
 
 class realExp : public literalExp {
@@ -527,6 +541,13 @@ public:
 
   types::ty *trans(coenv &e);
   types::ty *getType(coenv &) { return types::primReal(); }
+
+  template<typename T>
+  [[nodiscard]]
+  T getValue() const
+  {
+    return static_cast<T>(value);
+  }
 };
 
 
@@ -656,6 +677,7 @@ struct argument {
 #endif
 
   void prettyprint(ostream &out, Int indent);
+  void createSymMap(AsymptoteLsp::SymbolContext* symContext);
 };
 
 class arglist : public gc {
@@ -725,6 +747,8 @@ public:
   virtual argument& getRest() {
     return rest;
   }
+
+  virtual void createSymMap(AsymptoteLsp::SymbolContext* symContext);
 };
 
 // callExp has a global cache of resolved overloaded functions.  This clears
@@ -796,10 +820,20 @@ public:
     args->add(arg3);
   }
 
-  void prettyprint(ostream &out, Int indent);
+  void prettyprint(ostream &out, Int indent) override;
+  void createSymMap(AsymptoteLsp::SymbolContext* symContext) override;
 
-  types::ty *trans(coenv &e);
-  types::ty *getType(coenv &e);
+  using colorInfo = std::tuple<double, double, double>;
+
+  /**
+   * @return nullopt if callExp is not a color, pair<color, nullopt> if color is RGB,
+   * and pair<color, alpha> if color is RGBA.
+   */
+  optional<std::tuple<colorInfo, optional<double>,
+    AsymptoteLsp::posInFile, AsymptoteLsp::posInFile>> getColorInformation();
+
+  types::ty *trans(coenv &e) override;
+  types::ty *getType(coenv &e) override;
 
   // Returns true if the function call resolves uniquely without error.  Used
   // in implementing the special == and != operators for functions.
@@ -862,10 +896,12 @@ public:
   castExp(position pos, ty *target, exp *castee)
     : exp(pos), target(target), castee(castee) {}
 
-  void prettyprint(ostream &out, Int indent);
+  void prettyprint(ostream &out, Int indent) override;
 
-  types::ty *trans(coenv &e);
-  types::ty *getType(coenv &e);
+  types::ty *trans(coenv &e) override;
+  types::ty *getType(coenv &e) override;
+
+  void createSymMap(AsymptoteLsp::SymbolContext* symContext) override;
 };
 
 class nullaryExp : public callExp {
@@ -956,10 +992,16 @@ public:
   andOrExp(position pos, exp *left, symbol op, exp *right)
     : exp(pos), left(left), op(op), right(right) {}
 
-  virtual types::ty *trans(coenv &e) = 0;
-  virtual types::ty *getType(coenv &) {
+  virtual types::ty *trans(coenv &e) override = 0;
+  virtual types::ty *getType(coenv &) override {
     return types::primBoolean();
   }
+
+  void createSymMap(AsymptoteLsp::SymbolContext* symContext) override
+  {
+    left->createSymMap(symContext);
+    right->createSymMap(symContext);
+  }
 };
 
 class orExp : public andOrExp {
@@ -1034,20 +1076,22 @@ public:
   assignExp(position pos, exp *dest, exp *value)
     : exp(pos), dest(dest), value(value) {}
 
-  void prettyprint(ostream &out, Int indent);
+  void prettyprint(ostream &out, Int indent) override;
 
   // Don't write the result of an assignment to the prompt.
-  bool writtenToPrompt() { return false; }
+  bool writtenToPrompt() override { return false; }
 
-  void transAsType(coenv &e, types::ty *target);
-  types::ty *trans(coenv &e);
-  types::ty *getType(coenv &e);
+  void transAsType(coenv &e, types::ty *target) override;
+  types::ty *trans(coenv &e) override;
+  types::ty *getType(coenv &e) override;
+
+  void createSymMap(AsymptoteLsp::SymbolContext* symContext) override;
 };
 
 class selfExp : public assignExp {
   symbol op;
 
-  exp *ultimateValue(exp *dest) {
+  exp *ultimateValue(exp *dest) override {
     return new binaryExp(getPos(), dest, op, value);
   }
 
@@ -1055,9 +1099,9 @@ public:
   selfExp(position pos, exp *dest, symbol op, exp *value)
     : assignExp(pos, dest, value), op(op) {}
 
-  void prettyprint(ostream &out, Int indent);
+  void prettyprint(ostream &out, Int indent) override;
 
-  void transAsType(coenv &e, types::ty *target);
+  void transAsType(coenv &e, types::ty *target) override;
 };
 
 class prefixExp : public exp {
@@ -1068,15 +1112,15 @@ public:
   prefixExp(position pos, exp *dest, symbol op)
     : exp(pos), dest(dest), op(op) {}
 
-  void prettyprint(ostream &out, Int indent);
+  void prettyprint(ostream &out, Int indent) override;
 
-  bool scalable() { return false; }
+  bool scalable() override { return false; }
 
   // Don't write the result to the prompt.
-  bool writtenToPrompt() { return false; }
+  bool writtenToPrompt() override { return false; }
 
-  types::ty *trans(coenv &e);
-  types::ty *getType(coenv &e);
+  types::ty *trans(coenv &e) override;
+  types::ty *getType(coenv &e) override;
 };
 
 // Postfix expresions are illegal. This is caught here as we can give a
